--------------------------------------------------------
------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2015, by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- Frank Singhoff, Lab-STICC UMR 6285 laboratory, Université de Bretagne
--Occidentale
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- The full list of contributors and sponsors are given into AUTHORS.txt and
--SPONSORS.txt
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
------------------------------------------------------------------------------
-- Last update :
--    $Rev: 1249 $
--    $Date: 2014-08-28 07:02:15 +0200 (Fri, 28 Aug 2014) $
--    $Author: singhoff $
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Strings.Unbounded;  use Ada.Strings.Unbounded;
with unbounded_strings;      use unbounded_strings;
with GNAT.Command_Line;      use GNAT.Command_Line;
with GNAT.OS_Lib;            use GNAT.OS_Lib;
with Text_IO;                use Text_IO;
with Version;                use Version;
with Ada.IO_Exceptions;      use Ada.IO_Exceptions;
with GNAT.Current_Exception; use GNAT.Current_Exception;
with Call_Framework;         use Call_Framework;
with Ocarina.Configuration;
with Ocarina.AADL.Parser;
with Ocarina.AADL.Printer;
with Ocarina.Analyzer;       use Ocarina.Analyzer;
with Ocarina.AADL;           use Ocarina.AADL;
with Types;                  use Types;
with Ocarina.Entities;       use Ocarina.Entities;
with Ocarina.Parser;

procedure Aadl2aadl is

   File_Name : Unbounded_String := empty_string;

   procedure Usage is
   begin
      Put_Line
        ("aadl2aadl is a program which displays an aadl file to the screen.");

      New_Line;
      Put_Line(
"Check Cheddar home page for details :  http://beru.univ-brest.fr/~singhoff/cheddar "
);
      New_Line;
      New_Line;
      Put_Line ("Usage : aadl2aadl [switch] aadl-file-to-display");
      Put_Line ("   switch can be :");
      Put_Line ("            -u get this help");
      New_Line;
   end Usage;

   Root    : Node_Id := No_Node;
   Success : Boolean := False;

begin

   Copyright ("aadl2aadl");

   -- Get arguments
   --
   loop
      case Getopt ("u v ") is
         when ASCII.NUL =>
            exit;

         when 'u' =>
            Usage;
            OS_Exit (0);

         when others =>
            Usage;
            OS_Exit (0);
      end case;
   end loop;

   loop
      declare
         S : constant String := Get_Argument (Do_Expansion => True);

      begin
         exit when S'Length = 0;
         File_Name := File_Name & S;
      end;
   end loop;

   Call_Framework.initialize (False);
   Ocarina.Initialize;
   Ocarina.Configuration.Init_Modules;

   --  Parse AADL property files
   --
   Root := Ocarina.Parser.Parse ("AADL_Properties.aadl", Root);

   Root := Ocarina.Parser.Parse ("AADL_Project.aadl", Root);

   --  If the two files above have not been found, use those of
   --  Ocarina.
   --
   Root := Ocarina.Parser.Parse_Standard_Property_Sets (Root);

   if No (Root) then
      Put_Line ("aadl2aadl.adb : Cannot parse standard property sets");
   end if;

   Root := Ocarina.Parser.Parse ("Cheddar_Properties.aadl", Root);
   Root :=
      Ocarina.Parser.Parse ("User_Defined_Cheddar_Properties.aadl", Root);

   -- Read and Parse file
   --
   Root := Ocarina.Parser.Parse (To_String (File_Name), Root);

   Root := Ocarina.AADL.Parser.Process (To_String (File_Name), Root);

   --  Run AADL AST analysis
   --
   if Root /= No_Node then
      Success :=
         Ocarina.Analyzer.Analyze_Tree (Root, Default_Analyzer_Options);

      if not Success then
         Root := No_Node;
      end if;
   end if;

   --  Check that the AADL specification is OK
   --
   if Root = No_Node then
      Put_Line ("aadl2aadl.adb : Syntax error in the AADL specification");
   else
      Ocarina.AADL.Printer.Print_Subtree (Root);
   end if;

exception

   when GNAT.Command_Line.Invalid_Switch =>
      begin
         Put_Line ("aadl2aadl.adb : invalid Switch ; " & Full_Switch);
         Usage;
      end;
   when GNAT.Command_Line.Invalid_Parameter =>
      begin
         Put_Line
           ("aadl2aadl.adb : missing parameter for switch ; " & Full_Switch);
         Usage;
      end;

   when Ada.IO_Exceptions.Name_Error =>
      Put_Line ("aadl2aadl.adb : Can not open project files, Name_Error");
   when Ada.IO_Exceptions.Status_Error =>
      Put_Line ("aadl2aadl.adb : Can not open project files, Status_Error");
   when Ada.IO_Exceptions.Mode_Error =>
      Put_Line ("aadl2aadl.adb : Can not open project files, Mode_Error");
   when Ada.IO_Exceptions.Use_Error =>
      Put_Line ("aadl2aadl.adb : Can not open project files, Use_Error");
   when Ada.IO_Exceptions.Device_Error =>
      Put_Line ("aadl2aadl.adb : Can not open project files, Device_Error");
   when Ada.IO_Exceptions.End_Error =>
      Put_Line ("aadl2aadl.adb : Can not open project files, End_Error");
   when Ada.IO_Exceptions.Data_Error =>
      Put_Line ("aadl2aadl.adb : Can not open project files, Data_Error");
   when Ada.IO_Exceptions.Layout_Error =>
      Put_Line ("aadl2aadl.adb : Can not open project files, Layout_Error");

   when others =>
      Put_Line ("This is an internal aadl2aadl bug ... sorry");
      Put_Line ("Exception name : " & Exception_Name);
      Put_Line ("Exception message : " & Exception_Message);
      Put_Line ("Please, send a bug report to cheddar@listes.univ-brest.fr");
      Put_Line ("Do not forget to join the AADL files with your bug report");

end Aadl2aadl;
