------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
-----------------------------------------------------------------------------
-- Last update :
--    $Rev: 523 $
--    $Date: 2012-09-26 15:09:39 +0200 (Wed, 26 Sep 2012) $
--    $Author: fotsing $
------------------------------------------------------------------------------
------------------------------------------------------------------------------
with Framework_Config;                  use Framework_Config;
with Xml_Tag;                           use Xml_Tag;
with Scheduler;                         use Scheduler;
with Translate;                         use Translate;
with Tasks;                             use Tasks;
with Task_Set;                          use Task_Set;
use Task_Set.Generic_Task_Set;
with Scheduling_Analysis;               use Scheduling_Analysis;
use Scheduling_Analysis.Double_Tasks_Parameters_Package;
with unbounded_strings;                 use unbounded_strings;
with Dependency_Services;               use Dependency_Services;
with Multiprocessor_Services_Interface; use Multiprocessor_Services_Interface;
use Multiprocessor_Services_Interface.Scheduling_Result_Per_Processor_Package;
with Debug;                             use Debug;

package body Call_Dependency_Framework is

   procedure Chetto_Parameter_Modifications
     (Sys                : in out System;
      Result             : in out Unbounded_String;
      A_Processor        : in Generic_Processor_Ptr;
      Compute_Deadlines  : in Boolean;
      Compute_Priorities : in Boolean;
      Update_Tasks_Set   : in Boolean;
      Output             : in Output_Format := String_Output)
   is

      Tmp       : Tasks_Set;
      A_Task    : Generic_Task_Ptr;
      Iterator1 : Tasks_Iterator;

   begin

      Put_Debug ("Call Chetto_Parameter_Modificiations");

      if Compute_Deadlines then

         Result :=
           Result &
           unbounded_lf &
           Lb_Minus &
           Lb_Chetto_Deadline (Current_Language) &
           To_Unbounded_String (" (") &
           Lb_See (Current_Language) &
           To_Unbounded_String ("[5], [6]).") &
           unbounded_lf &
           unbounded_lf;

      else
         Result :=
           Result &
           unbounded_lf &
           Lb_Minus &
           Lb_Chetto_Priority (Current_Language) &
           To_Unbounded_String (" (") &
           Lb_See (Current_Language) &
           To_Unbounded_String ("[5], [6]).") &
           unbounded_lf &
           unbounded_lf;
      end if;

      -- If we must update parameters, we do not work on temporary tasks_set
      -- Call logic and display new parameters
      --
      if Update_Tasks_Set then
         Set_Parameters_According_To_Chetto
           (Sys.Dependencies,
            Sys.Tasks,
            Compute_Deadlines,
            Compute_Priorities);

         -- Display priorities/deadlines
         --
         reset_iterator (Sys.Tasks, Iterator1);
         loop

            current_element (Sys.Tasks, A_Task, Iterator1);
            Result := Result &
                      To_Unbounded_String ("- ") &
                      A_Task.name &
                      To_Unbounded_String (" => ");
            if Compute_Deadlines then
               Result := Result &
                         To_Unbounded_String (A_Task.deadline'Img) &
                         unbounded_lf;
            else
               Result := Result &
                         To_Unbounded_String (A_Task.priority'Img) &
                         unbounded_lf;
            end if;

            exit when is_last_element (Sys.Tasks, Iterator1);

            next_element (Sys.Tasks, Iterator1);
         end loop;

         if Compute_Priorities then
            Result := Result & Lb_Priorities_Are_Updated (Current_Language);
         else
            Result := Result & Lb_Deadlines_Are_Updated (Current_Language);
         end if;
         Result := Result & unbounded_lf & unbounded_lf & unbounded_lf;

      else
         duplicate (Sys.Tasks, Tmp);
         Set_Parameters_According_To_Chetto
           (Sys.Dependencies,
            Tmp,
            Compute_Deadlines,
            Compute_Priorities);

         -- Display priorities/deadlines
         --
         reset_iterator (Tmp, Iterator1);
         loop

            current_element (Tmp, A_Task, Iterator1);
            Result := Result &
                      To_Unbounded_String ("- ") &
                      A_Task.name &
                      To_Unbounded_String (" => ");
            if Compute_Deadlines then
               Result := Result &
                         To_Unbounded_String (A_Task.deadline'Img) &
                         unbounded_lf;
            else
               Result := Result &
                         To_Unbounded_String (A_Task.priority'Img) &
                         unbounded_lf;
            end if;

            exit when is_last_element (Tmp, Iterator1);

            next_element (Tmp, Iterator1);
         end loop;

         free (Tmp);

      end if;

      Result := Result & unbounded_lf;

   exception
      when Dependency_Services.Dependencies_Period_Error =>
         Result := Result &
                   Lb_Chetto_Error2 (Current_Language) &
                   unbounded_lf;
      when Dependency_Services.Cyclic_Graph_Error =>
         Result := Result &
                   Lb_Chetto_Error1 (Current_Language) &
                   unbounded_lf;
      when Dependency_Services.Task_Model_Error =>
         Result := Result &
                   Lb_Chetto_Error3 (Current_Language) &
                   unbounded_lf;
      when Dependency_Services.Deadline_Exhausted =>
         Result := Result &
                   Lb_Chetto_Error4 (Current_Language) &
                   unbounded_lf;

   end Chetto_Parameter_Modifications;




   procedure Compute_End_To_End_Response_Time
     (Sys              : in out System;
      Result           : in out Unbounded_String;
      Update_Tasks_Set : in Boolean;
      One_Step         : in Boolean;
      Output           : in Output_Format := String_Output)
   is

      Rt          : Response_Time_Table;
      Referencies : Unbounded_String := empty_string;
      Message     : Unbounded_String := empty_string;

   begin

      Put_Debug ("Call Compute_End_To_End_Response_Time");

      initialize (Rt);
      if Output = Xml_Output then
         Set_Tag;
      else
         Set_Empty;
      end if;

      begin

         Compute_End_To_End_Response_Time
           (Sys.Dependencies,
            Sys.Processors,
            Sys.Tasks,
            One_Step,
            Update_Tasks_Set,
            Referencies,
            Rt);

         if One_Step then
            Message := Message &
                       Start_Line &
                       Lb_Minus &
                       Lb_Jitter_From_Response_Time_One_Step (Current_Language)
                       &
                       Start_Ref;
         else
            Message := Message &
                       Start_Line &
                       Lb_Minus &
                       Lb_Jitter_From_Response_Time_All_Steps (
              Current_Language) &
                       Start_Ref;
         end if;

         Message := Message &
                    Referencies &
                    End_Ref &
                    End_Line &
                    unbounded_lf;

         if Update_Tasks_Set then
            Message := Message &
                       Start_Line &
                       Lb_Jitters_Are_Updated (Current_Language) &
                       End_Line &
                       unbounded_lf;
         end if;

         Message := Message &
                    Start_Line &
                    Lb_Task_Response_Time (Current_Language) &
                    unbounded_lf;

         for I in 0 .. Rt.nb_entries - 1 loop
            if (Rt.entries (I).item /= null) then
               Message :=
                 Message &
                 Start_Task &
                 To_Unbounded_String ("    ") &
                 Rt.entries (I).item.name &
                 To_Unbounded_String (" => ") &
                 Integer'Image (Integer (Rt.entries (I).data)) &
                 End_Task &
                 Lb_Comma &
                 Start_Jit &
                 Lb_Jitter (Current_Language) &
                 To_Unbounded_String ("=") &
                 Periodic_Task_Ptr (Rt.entries (I).item).jitter'Img &
                 To_Unbounded_String (" ") &
                 End_Jit;
               if (Natural (Rt.entries (I).data) >
                   Rt.entries (I).item.deadline)
               then
                  Message := Message &
                             Lb_Comma &
                             Lb_Check_Deadline (Current_Language) &
                             Rt.entries (I).item.deadline'Img;
                  Message := Message & To_Unbounded_String (")");
               end if;

               Message := Message & unbounded_lf;
            end if;
         end loop;
         Message := Message & End_Line & unbounded_lf;
         Message := Message & End_End_To_End & unbounded_lf;

      exception
         when dependency_services.Dependencies_Period_Error =>
            Message := Message & 
		       Lb_holistique_Error1(Current_Language) &
                       unbounded_lf;
      when Dependency_Services.Task_Model_Error =>
         Result := Result &
                   Lb_holistique_Error2(Current_Language) &
                   unbounded_lf;
         when Task_Set.Task_Must_Be_Periodic =>
            Message := Message &
                       Lb_Compute_Scheduling_Error_5 (Current_Language) &
                       unbounded_lf;
         when Task_Set.Task_Must_Have_Period_Equal_To_Deadline =>
            Message := Message &
                       Lb_Compute_Scheduling_Error_4 (Current_Language) &
                       unbounded_lf;
         when Task_Set.Task_Model_Error =>
            Message := Message &
                       Lb_Compute_Scheduling_Error_6 (Current_Language) &
                       unbounded_lf;
         when Processor_Utilization_Exceeded =>
            Message := Message &
                       Lb_Compute_Scheduling_Error_7 (Current_Language) &
                       unbounded_lf;
         when Task_Set.Start_Time_Error =>
            Message := Message &
                       Lb_Compute_Scheduling_Error_12 (Current_Language) &
                       unbounded_lf;
         when Task_Set.Offset_Error =>
            Message := Message &
                       Lb_Compute_Scheduling_Error_13 (Current_Language) &
                       unbounded_lf;
         when invalid_scheduler =>
            Message := Message &
                       Lb_Compute_Scheduling_Error_18(Current_Language) &
                       unbounded_lf;

      end;

      Result := Result & Message;

   end Compute_End_To_End_Response_Time;

end Call_Dependency_Framework;
