------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
-----------------------------------------------------------------------------
--
--
--
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.Strings.Unbounded; 		use Ada.Strings.Unbounded;
with unbounded_strings;     		use unbounded_strings;
with Framework_Config;      		use Framework_Config;
with Caches;                		use Caches;
with Cache_Access_Profile_Set; 		use Cache_Access_Profile_Set;
with sets;


package Cache_Set is

   no_cache_blocks_table 	: Cache_Blocks_Table;
   no_cache_access_profile  	: Cache_Access_Profiles_Set;

   ----------------------------------------------------------
   -- CACHE
   ----------------------------------------------------------

   package Generic_Cache_Set is new Sets (max_element    => Framework_Config.Max_Caches,
                                          element        => Generic_Cache_Ptr,
                                          free           => Free,
                                          copy           => Copy,
                                          put            => Put,
                                          xml_string     => XML_String,
                                          xml_ref_string => XML_Ref_String);

   use Generic_Cache_Set;

   type Caches_Set is new Generic_Cache_Set.set with private;

   subtype Caches_Range is Generic_Cache_Set.element_range;
   subtype Caches_Iterator is Generic_Cache_Set.iterator;

   -- Raised when a given cache is not in a cache unit
   -- set
   --
   Cache_Not_Found : exception;

   -- Raised when parameters provided to Add_cache are wrong
   --
   Invalid_Parameter : exception;

   ----------------------------------------------------------
   -- I/O operations
   ----------------------------------------------------------
   function Export_Aadl_Implementations
     (my_caches	    : in Caches_Set)
      return          Unbounded_String;

   function Export_Aadl_Declarations
     (my_caches     : in Caches_Set;
      Number_Of_Ht  : in Natural)
      return          Unbounded_String;

   ----------------------------------------------------------
   -- Procedure to proceed modification on a set
   ----------------------------------------------------------

   procedure Check_Cache
     (name           		: in Unbounded_String;
      cache_size 		: in Natural;
      line_size 		: in Natural;
      associativity 		: in Natural;
      block_reload_time 	: in Natural;
      coherence_protocol  	: in Cache_Coherence_Protocol_Type;
      replacement_policy 	: in Cache_Replacement_Policy_Type;
      cache_category		: in Cache_Type);

   -- Add a single cache to the set
   --
   procedure Add_Cache
     (my_caches			: in out Caches_Set;
      a_cache			: in out Generic_Cache_Ptr;
      name           		: in Unbounded_String;
      cache_size 		: in Natural;
      line_size 		: in Natural;
      associativity 		: in Natural;
      block_reload_time 	: in Natural;
      coherence_protocol  	: in Cache_Coherence_Protocol_Type;
      replacement_policy 	: in Cache_Replacement_Policy_Type;
      cache_category		: in Cache_Type;
      a_cache_blocks_table 	: in Cache_Blocks_Table := no_cache_blocks_table);


   procedure Add_Cache
     (my_caches			: in out Caches_Set;
      name           		: in Unbounded_String;
      cache_size 		: in Natural;
      line_size 		: in Natural;
      associativity 		: in Natural;
      block_reload_time 	: in Natural;
      coherence_protocol  	: in Cache_Coherence_Protocol_Type;
      replacement_policy 	: in Cache_Replacement_Policy_Type;
      cache_category		: in Cache_Type;
      a_cache_blocks_table 	: in Cache_Blocks_Table := no_cache_blocks_table);


   -- Update a cache in the set
   --
   procedure Update_Cache
     (my_caches			: in out Caches_Set;
      name           		: in Unbounded_String;
      cache_size 		: in Natural;
      line_size 		: in Natural;
      associativity 		: in Natural;
      block_reload_time 	: in Natural;
      coherence_protocol  	: in Cache_Coherence_Protocol_Type;
      replacement_policy 	: in Cache_Replacement_Policy_Type;
      cache_category		: in Cache_Type);


   -- Delete a cache in the set
   --

   procedure Delete_Cache
     (my_caches			: in out Caches_Set;
      name 			: in Unbounded_String);

   ------------------------------------------------------
   -- Read information from a set
   ------------------------------------------------------

   -- Get the cache pointer of the cache named "Name"
   --
   function Search_Cache
     (my_caches 	: in Caches_Set;
      name          	: in Unbounded_String)
      return          Generic_Cache_Ptr;
   function Search_Cache_by_id
     (my_caches 	: in Caches_Set;
      id          	: in Unbounded_String)
      return          Generic_Cache_Ptr;


private
   type Caches_Set is new Generic_Cache_Set.set with null record;

end Cache_Set;

