with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with unbounded_strings;     use unbounded_strings;
with Framework_Config;      use Framework_Config;
with CFG_Nodes;		    use CFG_Nodes;

with sets;


package CFG_Node_Set is

   ----------------------------------------------------
   -- In this package, we work with both CFG_Node_Table and CFG_Node_Set.
   -- The reason for that is because:
   -- + In the system, we use CFG_Node_Set
   -- + In the task, we use CFG_Node_Table
   -- We did by this way because this is the current implementation for other
   -- models of Cheddars such as processor and core_unit.
   ----------------------------------------------------


   ----------------------------------------------------------
   -- Definition of basic block set
   ----------------------------------------------------------

   package CFG_Node_Set is new Sets (
                                     max_element    => Framework_Config.Max_CFG_Nodes,
                                     element        => CFG_Node_Ptr,
                                     free           => Free,
                                     copy           => Copy,
                                     put            => Put,
                                     xml_string     => XML_String,
                                     xml_ref_string => XML_Ref_String);

   use CFG_Node_Set;

   type CFG_Nodes_Set is new CFG_Node_Set.set with private;

   subtype CFG_Nodes_Range is CFG_Node_Set.element_range;
   subtype CFG_Nodes_Iterator is CFG_Node_Set.iterator;

   ----------------------------------------------------------
   -- Exceptions defined in the set package
   ----------------------------------------------------------

   -- Raised when a given CFG node is not in a CFG node set
   --
   CFG_Node_Not_Found : exception;

   -- Raised when parameters provided to Add_Basic_Block are wrong
   --
   Invalid_Parameter : exception;

   ----------------------------------------------------------
   -- I/O operations
   ----------------------------------------------------------
   --     function Export_Aadl_Implementations
   --       (my_basic_blocks	: in Basic_Blocks_Set)
   --        return 		Unbounded_String;
   --
   --     function Export_Aadl_Declarations
   --       (my_basic_blocks	: in Basic_Blocks_Set;
   --        Number_Of_Ht	: in Natural)
   --        return		Unbounded_String;

   ----------------------------------------------------
   -- CHECK
   ----------------------------------------------------

   procedure Check_CFG_Node
     (name			: in Unbounded_String;
      node_type			: in CFG_Node_Type;
      graph_type		: in CFG_Graph_Type);

   ----------------------------------------------------
   -- ADD
   ----------------------------------------------------
   procedure Add_CFG_Node
     (my_cfg_nodes		: in out CFG_Nodes_Set;
      a_cfg_node		: in out CFG_Node_Ptr;
      name			: in Unbounded_String;
      node_type			: in CFG_Node_Type;
      graph_type		: in CFG_Graph_Type);

   procedure Add_CFG_Node
     (my_cfg_nodes		: in out CFG_Nodes_Set;
      name			: in Unbounded_String;
      node_type			: in CFG_Node_Type;
      graph_type		: in CFG_Graph_Type);

   ------------------------------------------------------
   -- SEARCH
   ------------------------------------------------------
   function Search_CFG_Node
     (my_cfg_nodes	: in CFG_Nodes_Set;
      name          	: in Unbounded_String)
      return		CFG_Node_Ptr;
   --Search CFG node by name

   function Search_CFG_Node_By_Id
     (my_cfg_nodes 	: in CFG_Nodes_Set;
      id 		: in Unbounded_String)
      return            CFG_Node_Ptr;
   --Search CFG node by cheddar_private_id


   ------------------------------------------------------
   -- DELETE
   ------------------------------------------------------

   procedure Delete_CFG_Node
     (my_cfg_nodes		: in out CFG_Nodes_Set;
      name 			: in Unbounded_String);

   ------------------------------------------------------
   -- UPDATE
   ------------------------------------------------------

   procedure Update_CFG_Node
     (my_cfg_nodes		: in out CFG_Nodes_Set;
      name			: in Unbounded_String;
      node_type			: in CFG_Node_Type;
      graph_type		: in CFG_Graph_Type);

   ------------------------------------------------------
   -- NEXT, PREVIOUS
   ------------------------------------------------------
   procedure Add_Next_CFG_Node
     (my_cfg_nodes		: in out CFG_Nodes_Set;
      name			: in Unbounded_String;
      next_node_name		: in Unbounded_String);
   -- Add the CFG node with name="next_node_name : in" to
   -- the next_nodes of the CFG node with name="name : in".

   procedure Add_Next_CFG_Node_By_Id
     (my_cfg_nodes		: in out CFG_Nodes_Set;
      a_cfg_node		: in out CFG_Node_Ptr;
      next_node_id		: in Unbounded_String);
   -- Add the CFG node with name="next_node_name : in" to
   -- the next_nodes of the CFG node a_cfg_node.

   procedure Add_Previous_CFG_Node
     (my_cfg_nodes		: in out CFG_Nodes_Set;
      name			: in Unbounded_String;
      previous_node_name	: in Unbounded_String);
   -- Add the CFG node with name="next_node_name : in" to
   -- the previous_nodes of the CFG node with name="name : in".

   procedure Add_Previous_CFG_Node_By_Id
     (my_cfg_nodes		: in out CFG_Nodes_Set;
      a_cfg_node		: in out CFG_Node_Ptr;
      previous_node_id		: in Unbounded_String);
   -- Add the CFG node with name="next_node_name : in" to
   -- the previous_nodes of the CFG node a_cfg_node.

   --procedure Set_Previous_CFG_Nodes
   --  (my_cfg_nodes 		: in out CFG_Nodes_Set);
   --If all the next_nodes is set, this method is used to automatically
   --set the previous nodes.

private
   type CFG_Nodes_Set is new CFG_Node_Set.set with null record;

end CFG_Node_Set;
