------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
-----------------------------------------------------------------------------
-- Last update :
--    $Rev: 523 $
--    $Date: 2012-09-26 15:09:39 +0200 (Wed, 26 Sep 2012) $
--    $Author: fotsing $
------------------------------------------------------------------------------
------------------------------------------------------------------------------
with unbounded_strings; use unbounded_strings;
with Ada.Exceptions;    use Ada.Exceptions;
with Translate;         use Translate;
with initialize_framework; use initialize_framework;
with objects.extended; use objects.extended;


package body Network_Set is

   procedure Check_network
     (My_networks  : in Networks_Set;
      Name         : in Unbounded_String;
      Network_Type : in Networks_Type)
   is

   begin

      if Name = "" then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
                (Lb_Network_Name (Current_Language) &
                Lb_Mandatory (Current_Language)));
      end if;

   if not Is_A_Valid_Identifier (Name) then
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String
               (Lb_network (Current_Language) &
                " " &
                Name &
                " : " &
               Lb_network_Name (Current_Language) &
                Lb_Colon &
                Lb_Invalid_Identifier (Current_Language)));
      end if;

   end Check_network;


 procedure Add_Network
     (My_Networks  : in out Networks_Set;
      Name         : in Unbounded_String;
      Network_Type : in Networks_Type)
is
dummy : network_ptr;
begin
	add_network(my_networks, dummy, name, network_type);
end add_network;


   procedure Add_Network
     (My_Networks  : in out Networks_Set;
      a_network    :  out network_ptr;
      Name         : in Unbounded_String;
      Network_Type : in Networks_Type)
   is

      My_Iterator : iterator;

   begin

check_initialize;

      Check_network (My_Networks, Name, Network_Type);

      if (get_number_of_elements (My_Networks) > 0) then
         reset_iterator (My_Networks, My_Iterator);

         loop
            current_element (My_Networks, A_Network, My_Iterator);

            if (Name = A_Network.name) then
		Raise_Exception
                 (Invalid_Parameter'Identity,
                  To_String
                     (Lb_network (Current_Language) &
                      " " &
                      Name &
                      " : " &
                      Lb_network_Name (Current_Language) &
                      Lb_Already_Defined (Current_Language)));

            end if;

            exit when is_last_element (My_Networks, My_Iterator);

            next_element (My_Networks, My_Iterator);

         end loop;
      end if;

      A_Network := new Network;

      A_Network.name         := Name;
      A_Network.network_type := Network_Type;

      add (My_Networks, A_Network);

   exception
      when Generic_Network_Set.full_set =>
         Raise_Exception
           (Invalid_Parameter'Identity,
            To_String (Lb_Can_Not_Define_More_Networks (Current_Language)));

   end Add_Network;

   procedure Update_Network
     (My_Networks  : in out Networks_Set;
      Name         : in Unbounded_String;
      Network_Type : in Networks_Type)
   is

      A_Network : Network_Ptr;

   begin

      A_Network := Search_Network (My_Networks, Name);
      delete (My_Networks, A_Network);
      Add_Network (My_Networks, Name, Network_Type);

   end Update_Network;

   function Search_Network
     (My_Networks : in Networks_Set;
      Name        : in Unbounded_String)
      return        Network_Ptr
   is
      My_Iterator : iterator;
      A_Network   : Network_Ptr;
      Result      : Network_Ptr;

      Found : Boolean := False;

   begin


if not is_empty (My_Networks)
then
      reset_iterator (My_Networks, My_Iterator);

      loop
         current_element (My_Networks, A_Network, My_Iterator);

         if (A_Network.name = Name) then
            Found  := True;
            Result := A_Network;
         end if;

         exit when is_last_element (My_Networks, My_Iterator);

         next_element (My_Networks, My_Iterator);

      end loop;
end if;

      if not Found then
         Raise_Exception
           (network_not_found'Identity,
            to_string(lb_name(current_language) & " = " & name) );
      end if;


      return Result;

   end Search_Network;


   function Export_Aadl_Implementations
     (My_Networks : in Networks_Set)
      return        Unbounded_String
   is
      My_Iterator : iterator;
      A_Network   : Network_Ptr;

      Result : Unbounded_String := empty_string;

   begin

      if not is_empty (My_Networks) then

         reset_iterator (My_Networks, My_Iterator);

         loop
            current_element (My_Networks, A_Network, My_Iterator);
            exit when is_last_element (My_Networks, My_Iterator);

            Result := Result &
                      To_Unbounded_String
                         ("bus " & To_String (A_Network.name));
            Result := Result &
                      To_Unbounded_String
                         ("end " & To_String (A_Network.name) & ";");

            Result := Result &
                      To_Unbounded_String
                         ("bus implementation " &
                          To_String (A_Network.name) &
                          ".Impl");
            Result := Result &
                      To_Unbounded_String
                         ("end " & To_String (A_Network.name) & ".Impl");

            next_element (My_Networks, My_Iterator);

         end loop;
      end if;

      return Result;

   end Export_Aadl_Implementations;

   function Export_Aadl_Declarations
     (My_Networks  : in Networks_Set;
      Number_Of_Ht : in Natural)
      return         Unbounded_String
   is
      My_Iterator : iterator;
      A_Network   : Network_Ptr;

      Result : Unbounded_String := empty_string;

   begin

      if not is_empty (My_Networks) then

         reset_iterator (My_Networks, My_Iterator);

         loop
            current_element (My_Networks, A_Network, My_Iterator);
            exit when is_last_element (My_Networks, My_Iterator);

            for I in 1 .. Number_Of_Ht loop
               Result := Result & ASCII.HT;
            end loop;
            Result :=
              Result &
              To_Unbounded_String
                 ("instancied_" &
                  To_String (A_Network.name) &
                  " : bus " &
                  To_String (A_Network.name) &
                  ".Impl;");

            next_element (My_Networks, My_Iterator);

         end loop;
      end if;

      return Result;

   end Export_Aadl_Declarations;


end Network_Set;
