------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
-----------------------------------------------------------------------------
-- Last update :
--    $Rev: 523 $
--    $Date: 2012-09-26 15:09:39 +0200 (Wed, 26 Sep 2012) $
--    $Author: fotsing $
------------------------------------------------------------------------------
------------------------------------------------------------------------------
with Processor_Set;         use Processor_Set;
use Processor_Set.Generic_Processor_Set;
with Scheduling_Analysis;   use Scheduling_Analysis;
use Scheduling_Analysis.Double_Tasks_Parameters_Package;
with Task_Set;              use Task_Set;
use Task_Set.Generic_Task_Set;
with Task_Dependencies;     use Task_Dependencies;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Buffers;               use Buffers;
with Tasks;                 use Tasks;

package Dependency_Services is

   -- Raised when tasks in a DAG have differents
   -- value of period
   --
   Dependencies_Period_Error : exception;

   -- Raised when tasks in a DAG are of different types
   --
   Task_Model_Error : exception;

   -- Raised when a method do not permit a cyclic graph
   --
   Cyclic_Graph_Error : exception;

   -- Raised if periodic producer/consumer tasks are expected
   --
   Task_Must_Be_Periodic : exception;

   -- Raised when deadlines are two small to apply Chetto algorithm
   --
   Deadline_Exhausted : exception;



   -- Compute task parameters of a DAG organized tasks
   -- according to Chetto/Blazewicz method
   --
   procedure Set_Parameters_According_To_Chetto
     (My_Dependencies    : in Tasks_Dependencies_Ptr;
      My_Tasks           : in out Tasks_Set;
      Compute_Deadlines  : in Boolean;
      Compute_Priorities : in Boolean);


   -- Compute end to end response time by the Holistic
   -- method
   --
   procedure Compute_End_To_End_Response_Time
     (My_Dependencies  : in Tasks_Dependencies_Ptr;
      My_Processors    : in Processors_Set;
      My_Tasks         : in out Tasks_Set;
      One_Step         : in Boolean;
      Update_Tasks_Set : in Boolean;
      Msg              : in out Unbounded_String;
      Rt               : in out Response_Time_Table);


   -- Test is two tables of response time are equals
   --
   function Is_Equal
     (Rtn  : in Response_Time_Table;
      Rtn1 : in Response_Time_Table)
      return Boolean;


   -- Test is all deadlines are missed when we compute
   -- end to end response time
   -- In this case, we stop the computing
   --
   function All_Deadlines_Missed
     (Rtn  : in Response_Time_Table)
      return Boolean;


   -- Check that all task of a dependency graph have the same period
   -- If "Processor_Name" is an empty string, do this check
   -- on all processors
   --
   procedure Dependencies_Periods_Control
     (My_Tasks       : in Tasks_Set;
      My_Deps        : in Tasks_Dependencies_Ptr;
      Processor_Name : in Unbounded_String);


   -- Check that all task of a dependency graph have the same type
   -- If "Processor_Name" is an empty string, do this check
   -- on all processors
   --
   procedure Dependencies_Task_Models_Control
     (My_Tasks       : in Tasks_Set;
      My_Deps        : in Tasks_Dependencies_Ptr;
      Processor_Name : in Unbounded_String);


   -- Check if producers/consumers are periodic
   --
   procedure Periodic_Buffer_Control
     (My_Tasks : in Tasks_Set;
      A_Buffer : in Buffer_Ptr);

end Dependency_Services;
