------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
-----------------------------------------------------------------------------
-- Last update :
--    $Rev: 422 $
--    $Date: 2011-04-20 20:59:09 +0200 (mer., 20 avr. 2011) $
--    $Author: singhoff $
------------------------------------------------------------------------------
------------------------------------------------------------------------------
with Xml_Tag;                use Xml_Tag;
with double_util;            use double_util;
with Translate;              use Translate;
with unbounded_strings;      use unbounded_strings;
with Text_IO;                use Text_IO;
with Ada.Exceptions;         use Ada.Exceptions;
with GNAT.Current_Exception; use GNAT.Current_Exception;

package body Scheduler.Fixed_Priority.aperiodic_server is

   procedure Check_Before_Scheduling
     (My_Scheduler   : in aperiodic_server_Scheduler;
      My_Tasks       : in Tasks_Set;
      Processor_Name : in Unbounded_String)
   is

   begin

      if My_Scheduler.parameters.period <= 0 then
         Raise_Exception
           (Invalid_scheduler_parameter'Identity,
            "Invalid period value");
      end if;

      if My_Scheduler.parameters.capacity <= 0 then
         Raise_Exception
           (Invalid_scheduler_parameter'Identity,
            "Invalid capacity value");
      end if;

   end Check_Before_Scheduling;



   procedure Do_Election
     (My_Scheduler       : in out aperiodic_server_Scheduler;
      Si                 : in out Scheduling_Information;
      Result             : in out Scheduling_Sequence_Ptr;
      Msg                : in out Unbounded_String;
      Current_Time       : in Natural;
      Processor_Name     : in Unbounded_String;
      Address_Space_Name : in Unbounded_String;
      My_Dependencies    : in Tasks_Dependencies_Ptr;
      With_Offsets       : in Boolean;
      With_Precedencies  : in Boolean;
      With_Resources     : in Boolean;
      With_jitters     : in Boolean;
      With_minimize_preemption     : in Boolean;
      Event_To_Generate  : in Time_Unit_Event_Type_Boolean_Table;
      Elected            : in out Tasks_Range;
      No_Task            : in out Boolean)
   is

      I : Tasks_Range := 0;

   begin

      My_Scheduler.periodic_Highest_Priority  := Natural'First;
      My_Scheduler.aperiodic_Highest_Priority := Natural'First;
      My_Scheduler.periodic_Elected            := 0;
      My_Scheduler.aperiodic_Elected           := 0;

      -- For each task, call "check_resources" to
      -- take care of priority modification (PCP and PIP)
      --
      loop

  if not Si.Tcbs (i).already_run_at_current_time then
         if (Si.Tcbs (I).Tsk.cpu_name = Processor_Name) and
            ((Address_Space_Name = To_Unbounded_String ("")) or
             (Address_Space_Name = Si.Tcbs (I).Tsk.address_space_name)) then
            if (Si.Tcbs (I).Wake_Up_Time <= Current_Time) and
               (Si.Tcbs (I).Rest_Of_Capacity /= 0)
            then
               Check_jitter(Si.Tcbs (I), Current_Time, Si.Tcbs(I).is_jitter_ready );
               if (With_jitters  = False) or (Si.Tcbs (I).is_jitter_ready)
               then

               if (With_Offsets = False) or
                  Check_Offset (Si.Tcbs (I), Current_Time)
               then
                  if (With_Precedencies = False) or
                     Check_Precedencies
                        (Si,
                         My_Dependencies,
                         Current_Time,
                         Si.Tcbs (I))
                  then
                     if With_Resources then
                        Check_Resource
                          (My_Scheduler,
                           Si,
                           Result,
                           Current_Time,
                           Si.Tcbs (I),
                           Fixed_Priority_Tcb_Ptr (Si.Tcbs (I)).Is_resource_Ready,
                           Event_To_Generate);
                     end if;
                  end if;
               end if;
               end if;
            end if;
            end if;
         end if;

         I := I + 1;
         exit when Si.Tcbs (I) = null;
      end loop;

      --  Look for the Highest periodic/sporadic/poisson process ready
      --  priority task
      --
      I := 0;
      loop
         if ((Si.Tcbs (I).Tsk.task_type = Periodic_Type) or
             (Si.Tcbs (I).Tsk.task_type = Sporadic_Type) or
             (Si.Tcbs (I).Tsk.task_type = Poisson_Type))
         then
            if (Si.Tcbs (I).Tsk.cpu_name = Processor_Name) then
	     if Check_Core_Assignment(my_scheduler, Si.Tcbs (I))
               then
                  if (Si.Tcbs (I).Wake_Up_Time <= Current_Time) and
                     (Natural (
                       Fixed_Priority_Tcb_Ptr (Si.Tcbs (I)).Current_Priority) >
                      My_Scheduler.periodic_Highest_Priority) and
                     (Si.Tcbs (I).Rest_Of_Capacity /= 0)
                  then
               Check_jitter(Si.Tcbs (I), Current_Time, Si.Tcbs(I).is_jitter_ready );
               if (With_jitters  = False) or (Si.Tcbs (I).is_jitter_ready)
               then
                     if (With_Offsets = False) or
                        Check_Offset (Si.Tcbs (I), Current_Time)
                     then
                        if (With_Precedencies = False) or
                           Check_Precedencies
                              (Si,
                               My_Dependencies,
                               Current_Time,
                               Si.Tcbs (I))
                        then
                           if (With_Resources = False) or
                              Fixed_Priority_Tcb_Ptr (Si.Tcbs (I)).Is_resource_Ready
                           then

                              My_Scheduler.periodic_Highest_Priority :=
                                Natural (
                                Fixed_Priority_Tcb_Ptr (Si.Tcbs (I)).
                                Current_Priority);
                              My_Scheduler.periodic_Elected           := I;
                           end if;
                        end if;
                     end if;
                  end if;
                  end if;
               end if;
            end if;
         end if;

         I := I + 1;
         exit when Si.Tcbs (I) = null;
      end loop;

      -- Check if some aperiodic task exists and select the Highest priority
      -- aperiodic task
      --
      I := 0;
      loop
      if not Si.Tcbs (i).already_run_at_current_time then
         if (Si.Tcbs (I).Tsk.task_type = Aperiodic_Type) then
            if (Si.Tcbs (I).Tsk.cpu_name = Processor_Name) then
		if Check_Core_Assignment(my_scheduler, Si.Tcbs (I))
               	then
                  if (Si.Tcbs (I).Wake_Up_Time <= Current_Time) and
                     (Natural (
                       Fixed_Priority_Tcb_Ptr (Si.Tcbs (I)).Current_Priority) >
                      My_Scheduler.aperiodic_Highest_Priority) and
                     (Si.Tcbs (I).Rest_Of_Capacity /= 0)
                  then
               Check_jitter(Si.Tcbs (I), Current_Time, Si.Tcbs(I).is_jitter_ready );
               if (With_jitters  = False) or (Si.Tcbs (I).is_jitter_ready)
               then
                     if (With_Offsets = False) or
                        Check_Offset (Si.Tcbs (I), Current_Time)
                     then
                        if (With_Precedencies = False) or
                           Check_Precedencies
                              (Si,
                               My_Dependencies,
                               Current_Time,
                               Si.Tcbs (I))
                        then
                           if (With_Resources = False) or
                              Fixed_Priority_Tcb_Ptr (Si.Tcbs (I)).Is_resource_Ready
                           then

                              My_Scheduler.aperiodic_Highest_Priority :=
                                Natural (
                                Fixed_Priority_Tcb_Ptr (Si.Tcbs (I)).
                                Current_Priority);
                              My_Scheduler.aperiodic_Elected           := I;
                           end if;
                        end if;
                     end if;
                     end if;
                  end if;
               end if;
            end if;
         end if;
       end if;

         I := I + 1;
         exit when Si.Tcbs (I) = null;
      end loop;

   end Do_Election;

   procedure Specific_Scheduler_Initialization
     (My_Scheduler       : in out aperiodic_server_Scheduler;
      Si                 : in out Scheduling_Information;
      Processor_Name     : in Unbounded_String;
      address_space_name : in Unbounded_String;
      My_Tasks           : in out Tasks_Set;
      my_schedulers      : in Scheduler_table;
      My_Resources       : in out Resources_Set;
      My_Buffers         : in out Buffers_Set;
      My_Messages        : in Messages_Set;
      Msg                : in out Unbounded_String)
   is

   begin
      Specific_Scheduler_Initialization
        (Fixed_Priority_Scheduler (My_Scheduler),
         Si,
         Processor_Name,
         address_space_name,
         My_Tasks,
         my_schedulers,
         My_Resources,
         My_Buffers,
         My_Messages,
         Msg);
		 
		 My_Scheduler.current_server_wake_time :=
			My_Scheduler.parameters.start_time;
		 My_Scheduler.current_server_capacity :=
			My_Scheduler.parameters.capacity;

   end Specific_Scheduler_Initialization;

end Scheduler.Fixed_Priority.aperiodic_server;
