------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
-----------------------------------------------------------------------------
-- Last update :
--    $Rev: 523 $
--    $Date: 2012-09-26 15:09:39 +0200 (Wed, 26 Sep 2012) $
--    $Author: fotsing $
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with unbounded_strings; use unbounded_strings;

package Scheduler.Fixed_Priority is

   type Fixed_Priority_Scheduler is abstract new Generic_Scheduler with
      private;
   type Fixed_Priority_Scheduler_Ptr is access all Fixed_Priority_Scheduler'
     Class;

   type Fixed_Priority_Tcb is new Tcb with record

      -- With fixed priority, task priority may change due to resource accesses
      -- This attribute stores the current task priorities taking care of resource accesses
      Current_Priority : Priority_Range;

      -- Store if a task is allowed to run because of its resource access requirement, i.e. does
      -- the task would be blocked now due to lock resources ??
      Is_Resource_Ready         : Boolean;

   end record;


   type Fixed_Priority_Tcb_Ptr is access all Fixed_Priority_Tcb'Class;

   procedure Initialize (A_Tcb : in out Fixed_Priority_Tcb);

   ----------------------------------------

   type Fixed_Priority_Resource is new Shared_Resource with record
      Priority_Ceiling : Priority_Range;
   end record;

   type Fixed_Priority_Resource_Ptr is access all Fixed_Priority_Resource'
     Class;


--------------------------------------------------------------------------------------------------------------------------------------------
-- This text is a summary of the blocking and priority inheritance implemented in this package
--------------------------------------------------------------------------------------------------------------------------------------------
--   Protocol        | Conditions to allocate         	| action when requesting   	| action on allocate  	| action on release
--		     |					| task is blocked on		| 			| 
--		     |					| allocation request		| 			| 
--------------------------------------------------------------------------------------------------------------------------------------------
--   no_protocol     | resource.state>0 		|         			| resource.state--	| resource.state++
--------------------------------------------------------------------------------------------------------------------------------------------
--   PIP             | 			  		| task in critical section	|        		| task leaving
--	             |					| inherits requesting task	|        		| resource
--		     |					| priority			|        		| gets its initial priority
--                   | resource.state>0 		| 	 			| resource.state--	| resource.state++
--------------------------------------------------------------------------------------------------------------------------------------------
--   IPCP            | resource.state>0 		| 				| resource.state--	| resource.state++
--                   |                  		| 				| 			|
--                   |                  		| 				| requestion task priority = max (static priority,
--                   |                  		| 				|   ceiling priority of all allocated resources)
--------------------------------------------------------------------------------------------------------------------------------------------
--   PCP             | resource.state>0 AND		| task in critical section	| 			| task leaving
--                   | priority of requesting task is	| inherits requesting task	| 			| resource
--                   | greater than all priority ceiling| priority			| 			| gets its initial priority
--                   | of resource that are allocated	|				|			|
--                   | by other tasks			|				|			|
--                   |                      		| 	 			| resource.state--	| resource.state++
--------------------------------------------------------------------------------------------------------------------------------------------

   

   function Build_Resource
     (My_Scheduler : in Fixed_Priority_Scheduler;
      A_Resource   : Generic_Resource_Ptr)
      return         Shared_Resource_Ptr;

   function Build_Tcb
     (My_Scheduler : in Fixed_Priority_Scheduler;
      A_Task       : Generic_Task_Ptr)
      return         Tcb_Ptr;

   procedure Do_Election
     (My_Scheduler       : in out Fixed_Priority_Scheduler;
      Si                 : in out Scheduling_Information;
      Result             : in out Scheduling_Sequence_Ptr;
      Msg                : in out Unbounded_String;
      Current_Time       : in Natural;
      Processor_Name     : in Unbounded_String;
      Address_Space_Name : in Unbounded_String;
      My_Dependencies    : in Tasks_Dependencies_Ptr;
      With_Offsets       : Boolean;
      With_Precedencies  : Boolean;
      With_Resources     : Boolean;
      With_jitters     : in Boolean;
      With_minimize_preemption     : in Boolean;
      Event_To_Generate  : in Time_Unit_Event_Type_Boolean_Table;
      Elected            : in out Tasks_Range;
      No_Task            : in out Boolean);

   procedure Check_Resource
     (My_Scheduler      : in out Fixed_Priority_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Current_Time      : in Natural;
      A_Tcb             : in Tcb_Ptr;
      Is_Ready          : out Boolean;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table);

   procedure Allocate_Resource
     (My_Scheduler      : in out Fixed_Priority_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Current_Time      : in Natural;
      A_Tcb             : Tcb_Ptr;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table);

   procedure Release_Resource
     (My_Scheduler      : in out Fixed_Priority_Scheduler;
      Si                : in out Scheduling_Information;
      Result            : in out Scheduling_Sequence_Ptr;
      Current_Time      : in Natural;
      A_Tcb             : Tcb_Ptr;
      Event_To_Generate : in Time_Unit_Event_Type_Boolean_Table);

   procedure Specific_Scheduler_Initialization
     (My_Scheduler       : in out Fixed_Priority_Scheduler;
      Si                 : in out Scheduling_Information;
      Processor_Name     : in Unbounded_String;
      address_space_name : in Unbounded_String;
      My_Tasks           : in out Tasks_Set;
      my_schedulers      : in Scheduler_table;
      My_Resources       : in out Resources_Set;
      My_Buffers         : in out Buffers_Set;
      My_Messages        : in Messages_Set;
      Msg                : in out Unbounded_String);

   procedure Compute_Ceiling_Of_Resources
     (My_Scheduler       : in out Fixed_Priority_Scheduler;
      Si                 : in out Scheduling_Information;
      Processor_Name     : in Unbounded_String;
      Address_space_Name : in Unbounded_String;
      My_Tasks           : in out Tasks_Set;
      My_Resources       : in out Resources_Set);

   procedure Dispatched_Change_Current_Priority
     (My_Scheduler : in out Fixed_Priority_Scheduler;
      A_Tcb        : in Fixed_Priority_Tcb_Ptr;
      New_Priority : Priority_Range);

   procedure Change_Current_Priority
     (My_Scheduler : in out Fixed_Priority_Scheduler'Class;
      A_Tcb        : in Fixed_Priority_Tcb_Ptr;
      New_Priority : Priority_Range);

   function Produce_Running_Task_Event
     (My_Scheduler : in Fixed_Priority_Scheduler;
      A_Task       : in Tcb_Ptr)
      return         Time_Unit_Event_Ptr;

   function Export_Xml_Event_Running_Task
     (My_Scheduler : in Fixed_Priority_Scheduler;
      An_Event     : in Time_Unit_Event_Ptr)
      return         Unbounded_String;

private

   type Fixed_Priority_Scheduler is abstract new Generic_Scheduler with record
      Used_Resource : Boolean := False;
   end record;

end Scheduler.Fixed_Priority;
