----------------------------------
----------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
------------------------------------------------------------------------------
-- Last update :
--    $Rev: 97 $
--    $Date: 2007-07-20 15:17:15 +0200 (Fri, 20 Jul 2007) $
--    $Author: singhoff $
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Text_IO;                           use Text_IO;
with Processors;                        use Processors;
with Processor_Set;                     use Processor_Set;
use Processor_Set.Generic_Processor_Set;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with unbounded_strings;                 use unbounded_strings;
use unbounded_strings.strings_table_package;
use unbounded_strings.unbounded_string_list_package;
with Parameters;                        use Parameters;
with Parameters.extended;               use Parameters.extended;
use Parameters.Framework_Parameters_Table_Package;
with Task_Set;                          use Task_Set;
use Task_Set.Generic_Task_Set;
with Systems;                           use Systems;
with framework;                         use framework;
with Call_Framework;                    use Call_Framework;
with Call_Framework_Interface;          use Call_Framework_Interface;
use Call_Framework_Interface.Framework_Response_Package;
use Call_Framework_Interface.Framework_Request_Package;
with Call_Scheduling_Framework;         use Call_Scheduling_Framework;
with Multiprocessor_Services;           use Multiprocessor_Services;
with Multiprocessor_Services_Interface; use Multiprocessor_Services_Interface;
use Multiprocessor_Services_Interface.Scheduling_Result_Per_Processor_Package;

procedure Comprehensive_Call is

   procedure Call_Processor_Analysis
     (A_Sys : in out System;
      Name  : in Unbounded_String)
   is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;
      A_Param       : Parameter_Ptr;

   begin
      Initialize (Response_List);
      Initialize (Request_List);

      Initialize (A_Request);
      A_Request.target    := Name;
      A_Request.statement := Scheduling_Simulation_Preemption_Number;
      Add (Request_List, A_Request);

      Initialize (A_Request);
      A_Request.target    := Name;
      A_Request.statement := Scheduling_Simulation_Context_Switch_Number;
      Add (Request_List, A_Request);

      Initialize (A_Request);
      A_Param                := new Parameter (Boolean_Parameter);
      A_Param.parameter_name := To_Unbounded_String ("worst_case");
      A_Param.boolean_value  := True;
      Add (A_Request.param, A_Param);

      A_Param                := new Parameter (Boolean_Parameter);
      A_Param.parameter_name := To_Unbounded_String ("best_case");
      A_Param.boolean_value  := True;
      Add (A_Request.param, A_Param);

      A_Param                := new Parameter (Boolean_Parameter);
      A_Param.parameter_name := To_Unbounded_String ("average_case");
      A_Param.boolean_value  := True;
      Add (A_Request.param, A_Param);

      A_Request.target    := Name;
      A_Request.statement := Scheduling_Simulation_Response_Time;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (A_Sys, Request_List, Response_List);

      Put_Line (To_String (Response_List.Entries (0).title));
      for J in 0 .. Response_List.Nb_Entries - 1 loop
         Put_Line (To_String (Response_List.Entries (J).text));
      end loop;

   end Call_Processor_Analysis;

   procedure Run_Event_Analyzer (A_Sys : in out System) is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin
      Initialize (Response_List);
      Initialize (Request_List);

      Initialize (A_Request);
      A_Request.statement := Scheduling_Simulation_Run_Event_Handler;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (A_Sys, Request_List, Response_List);

      Put_Line (To_String (Response_List.Entries (0).title));
      for J in 0 .. Response_List.Nb_Entries - 1 loop
         Put_Line (To_String (Response_List.Entries (J).text));
      end loop;

   end Run_Event_Analyzer;

   procedure Compute_Scheduling (A_Sys : in out System) is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;
      A_Param       : Parameter_Ptr;

   begin

      Initialize (Response_List);
      Initialize (Request_List);
      Initialize (A_Request);
      A_Request.statement    := Scheduling_Simulation_Time_Line;
      A_Param                := new Parameter (Integer_Parameter);
      A_Param.parameter_name := To_Unbounded_String ("period");
      A_Param.integer_value  := 100;
      Add (A_Request.param, A_Param);
      A_Param                := new Parameter (Boolean_Parameter);
      A_Param.parameter_name := To_Unbounded_String ("schedule_with_offsets");
      A_Param.boolean_value  := True;
      Add (A_Request.param, A_Param);
      A_Param                := new Parameter (Boolean_Parameter);
      A_Param.parameter_name :=
         To_Unbounded_String ("schedule_with_precedencies");
      A_Param.boolean_value  := True;
      Add (A_Request.param, A_Param);
      A_Param                := new Parameter (Boolean_Parameter);
      A_Param.parameter_name :=
         To_Unbounded_String ("schedule_with_resources");
      A_Param.boolean_value  := True;
      Add (A_Request.param, A_Param);
      A_Param                := new Parameter (Integer_Parameter);
      A_Param.parameter_name := To_Unbounded_String ("seed_value");
      A_Param.integer_value  := 0;
      Add (A_Request.param, A_Param);
      Add (Request_List, A_Request);
      Sequential_Framework_Request (A_Sys, Request_List, Response_List);

      Put_Line (To_String (Response_List.Entries (0).title));
      for J in 0 .. Response_List.Nb_Entries - 1 loop
         Put_Line (To_String (Response_List.Entries (J).text));
      end loop;

   end Compute_Scheduling;

   procedure Run_Basic_Feasibility_Tests (A_Sys : in out System) is

      Response_List : Framework_Response_Table;
      Request_List  : Framework_Request_Table;
      A_Request     : Framework_Request;

   begin

      Initialize (Response_List);
      Initialize (Request_List);
      Initialize (A_Request);
      A_Request.statement := Scheduling_Feasibility_Basics;
      Add (Request_List, A_Request);

      Sequential_Framework_Request (A_Sys, Request_List, Response_List);

      Put_Line (To_String (Response_List.Entries (0).title));
      for J in 0 .. Response_List.Nb_Entries - 1 loop
         Put_Line (To_String (Response_List.Entries (J).text));
      end loop;

   end Run_Basic_Feasibility_Tests;

   Sys : System;

   Project_File_List     : unbounded_string_list;
   A_Processor           : Generic_Processor_Ptr;
   My_Processor_Iterator : Processors_Iterator;
   Project_File_Dir_List : unbounded_string_list;

begin

   -- Initialize the Cheddar framework
   --
   Call_Framework.initialize (False);

   -- Read project file
   --
   initialize (Project_File_List);
   Systems.Read_From_Xml_File
     (Sys,
      Project_File_Dir_List,
      To_Unbounded_String ("arinc653.xml"));

   -- Compute basic feasibility tests
   --
   Run_Basic_Feasibility_Tests (Sys);

   -- Compute the scheduling
   --
   Compute_Scheduling (Sys);

   -- Perform analysis on the computed scheduling for each processor
   --
   reset_iterator (Sys.Processors, My_Processor_Iterator);
   loop
      current_element (Sys.Processors, A_Processor, My_Processor_Iterator);

      Call_Processor_Analysis (Sys, A_Processor.name);

      exit when is_last_element (Sys.Processors, My_Processor_Iterator);
      next_element (Sys.Processors, My_Processor_Iterator);
   end loop;

   -- Run event analyzer
   --
   Run_Event_Analyzer (Sys);

   -- Export Event table into the event_table.xml file
   --
   Write_To_Xml_File
     (framework.Sched,
      Sys,
      To_Unbounded_String ("event_table.xml"));

end Comprehensive_Call;
