------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
------------------------------------------------------------------------------
-- Last update :
--    $Rev: 97 $
--    $Date: 2007-07-20 15:17:15 +0200 (Fri, 20 Jul 2007) $
--    $Author: singhoff $
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with io_tools;                          use io_tools;
with Text_IO;                           use Text_IO;
with Processor_Set;                     use Processor_Set;
use Processor_Set.Generic_Processor_Set;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with unbounded_strings;                 use unbounded_strings;
use unbounded_strings.strings_table_package;
use unbounded_strings.unbounded_string_list_package;
with Parameters;                        use Parameters;
with Parameters.extended;               use Parameters.extended;
use Parameters.Framework_Parameters_Table_Package;
with Task_Set;                          use Task_Set;
use Task_Set.Generic_Task_Set;
with Systems;                           use Systems;
with Call_Framework;                    use Call_Framework;
with Call_Framework_Interface;          use Call_Framework_Interface;
use Call_Framework_Interface.Framework_Response_Package;
use Call_Framework_Interface.Framework_Request_Package;
with framework;                         use framework;
with Call_Scheduling_Framework;         use Call_Scheduling_Framework;
with Multiprocessor_Services;           use Multiprocessor_Services;
with Multiprocessor_Services_Interface; use Multiprocessor_Services_Interface;
use Multiprocessor_Services_Interface.Scheduling_Result_Per_Processor_Package;
with GNAT.Command_Line;
with GNAT.OS_Lib;                       use GNAT.OS_Lib;
with Version;                           use Version;
with Ada.Exceptions;                    use Ada.Exceptions;
with ARINC_653_Schema;                  use ARINC_653_Schema;
with arinc_653_Services;                use arinc_653_Services;

procedure produce_arinc_pst is

   -- A set of variables required to call the framework
   --
   Response_List         : Framework_Response_Table;
   Request_List          : Framework_Request_Table;
   A_Request             : Framework_Request;
   A_Param               : Parameter_Ptr;
   Sys                   : System;
   Project_File_List     : unbounded_string_list;
   Project_File_Dir_List : unbounded_string_list;

   -- Is the command should run in verbose mode ?
   --
   Verbose : Boolean := False;

   -- The XML file to read
   --
   Input_File      : Boolean := False;
   Input_File_Name : Unbounded_String;

   -- The XML file to write
   --
   Output_File      : Boolean := False;
   Output_File_Name : Unbounded_String;

   -- The duration on which we must compute the scheduling
   --
   Period_String : Unbounded_String;
   Period        : Natural;
   Ok            : Boolean;

   -- The ARINC module to produce from the Cheddar ADL model
   --
   a_module : ARINC_653_Module_Ptr;

   procedure Usage is
   begin
      Put_Line(
"produce_arinc_pst is a program which computes a scheduling from an XML  Cheddar project file and save the result into a second XML file which is compliant with ARINC 653 PST table."
);

      New_Line;
      Put_Line(
"Check Cheddar home page for details :  http://beru.univ-brest.fr/~singhoff/cheddar "
);
      New_Line;
      New_Line;
      Put_Line ("Usage : produce_arinc_pst [switch] period");
      Put_Line ("   switch can be :");
      Put_Line ("            -u get this help");
      Put_Line ("            -v verbose mode ");
      Put_Line(
"            -o file-name, write the PST into the XML file  file-name ");
      Put_Line(
"            -i file-name, read the system to analyze from the XML file  file-name "
);
      New_Line;
   end Usage;

begin

   Copyright ("produce_arinc_pst");

   -- Get arguments
   --
   loop
      case GNAT.Command_Line.Getopt ("u v i: o:") is
         when ASCII.NUL =>
            exit;

         when 'i' =>
            Input_File      := True;
            Input_File_Name :=
               To_Unbounded_String (GNAT.Command_Line.Parameter);

         when 'o' =>
            Output_File      := True;
            Output_File_Name :=
               To_Unbounded_String (GNAT.Command_Line.Parameter);
         when 'v' =>
            Verbose := True;
         when 'u' =>
            Usage;
            OS_Exit (0);

         when others =>
            Usage;
            OS_Exit (0);
      end case;
   end loop;

   loop
      declare
         S : constant String :=
           GNAT.Command_Line.Get_Argument (Do_Expansion => True);

      begin
         exit when S'Length = 0;
         Period_String := Period_String & S;
      end;
   end loop;

   -- Check the period on which we will compute the scheduling
   --
   to_natural (Period_String, Period, Ok);
   if not Ok then
      Raise_Exception
        (Constraint_Error'Identity,
         "The period to compute the scheduling must be a numeric value");
   end if;

   -- Is an input file and an output file given ???
   --
   if (not Input_File) or (not Output_File) then
      Usage;
      OS_Exit (0);
   end if;

   -- Initialize the Cheddar framework
   --
   Call_Framework.initialize (False);

   -- Read the XML project file
   --
   initialize (Project_File_List);
   Systems.Read_From_Xml_File (Sys, Project_File_Dir_List, Input_File_Name);

   -- Compute the scheduling on the period given by the argument
   --
   Initialize (Response_List);
   Initialize (Request_List);
   Initialize (A_Request);
   A_Request.statement    := Scheduling_Simulation_Time_Line;
   A_Param                := new Parameter (Integer_Parameter);
   A_Param.parameter_name := To_Unbounded_String ("period");
   A_Param.integer_value  := Period;
   Add (A_Request.param, A_Param);
   A_Param                := new Parameter (Boolean_Parameter);
   A_Param.parameter_name := To_Unbounded_String ("schedule_with_offsets");
   A_Param.boolean_value  := True;
   Add (A_Request.param, A_Param);
   A_Param                := new Parameter (Boolean_Parameter);
   A_Param.parameter_name :=
      To_Unbounded_String ("schedule_with_precedencies");
   A_Param.boolean_value  := True;
   Add (A_Request.param, A_Param);
   A_Param                := new Parameter (Boolean_Parameter);
   A_Param.parameter_name := To_Unbounded_String ("schedule_with_resources");
   A_Param.boolean_value  := True;
   Add (A_Request.param, A_Param);
   A_Param                := new Parameter (Integer_Parameter);
   A_Param.parameter_name := To_Unbounded_String ("seed_value");
   A_Param.integer_value  := 0;
   Add (A_Request.param, A_Param);
   Add (Request_List, A_Request);
   Sequential_Framework_Request (Sys, Request_List, Response_List);

   if Verbose then
      Put_Line (To_String (Response_List.Entries (0).title));
      for J in 0 .. Response_List.Nb_Entries - 1 loop
         Put_Line (To_String (Response_List.Entries (J).text));
      end loop;
   end if;

   -- Export PST
   --
   a_module := Build_ARINC653_Module (framework.Sched, Sys, Period);
   write_sequential_file (Output_File_Name, XML_String (a_module));

end produce_arinc_pst;
