------------------------------------------------------
--------------------------
------------------------------------------------------------------------------
-- Cheddar is a GNU GPL real time scheduling analysis tool.
-- This program provides services to automatically check performances
-- of real time architectures.
--
-- Copyright (C) 2002-2010,  by Frank Singhoff, Alain Plantec, Jerome Legrand
--
-- The Cheddar project was started in 2002 by
-- the LISyC Team, University of Western Britanny.
--
-- Since 2008, Ellidiss technologies also contributes to the development of
-- Cheddar and provides industrial support.
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : cheddar@listes.univ-brest.fr
--
------------------------------------------------------------------------------
-- Last update :
--    $Rev: 97 $
--    $Date: 2007-07-20 15:17:15 +0200 (Fri, 20 Jul 2007) $
--    $Author: singhoff $
------------------------------------------------------------------------------
------------------------------------------------------------------------------

with Ada.IO_Exceptions;                 use Ada.IO_Exceptions;
with GNAT.Current_Exception;            use GNAT.Current_Exception;
with Text_IO;                           use Text_IO;
with Ada.Strings.Unbounded;             use Ada.Strings.Unbounded;
with unbounded_strings;                 use unbounded_strings;
use unbounded_strings.strings_table_package;
use unbounded_strings.unbounded_string_list_package;
with unbounded_strings;                 use unbounded_strings;
with Parameters;                        use Parameters;
with Parameters.extended;               use Parameters.extended;
use Parameters.Framework_Parameters_Table_Package;
with Systems;                           use Systems;
with framework;                         use framework;
with Call_Framework;                    use Call_Framework;
with Call_Framework_Interface;          use Call_Framework_Interface;
use Call_Framework_Interface.Framework_Response_Package;
use Call_Framework_Interface.Framework_Request_Package;
with Call_Scheduling_Framework;         use Call_Scheduling_Framework;
with Multiprocessor_Services;           use Multiprocessor_Services;
with Multiprocessor_Services_Interface; use Multiprocessor_Services_Interface;
use Multiprocessor_Services_Interface.Scheduling_Result_Per_Processor_Package;
with GNAT.Command_Line;
with GNAT.OS_Lib;                       use GNAT.OS_Lib;
with Version;                           use Version;
with Ada.Exceptions;                    use Ada.Exceptions;
with Time_Unit_Events;                  use Time_Unit_Events;

with architecture_models; use architecture_models;

procedure schedule_hardcoded_model is

   -- A set of variables required to call the framework
   --
   Response_List         : Framework_Response_Table;
   Request_List          : Framework_Request_Table;
   A_Request             : Framework_Request;
   A_Param               : Parameter_Ptr;
   Sys                   : System;
   Project_File_List     : unbounded_string_list;
   Project_File_Dir_List : unbounded_string_list;

   -- Is the command should run in verbose mode ?
   --
   Verbose : Boolean := False;

   -- The duration on which we must compute the scheduling
   --
   Period_String : Unbounded_String;
   Period        : Natural;
   Ok            : Boolean;

   procedure Usage is
   begin
      Put_Line
        ("schedule_hardcoded_model is a program which computes a scheduling ");

      New_Line;
      Put_Line(
"Check Cheddar home page for details :  http://beru.univ-brest.fr/~singhoff/cheddar "
);
      New_Line;
      New_Line;
      Put_Line ("Usage : schedule_hardcoded_model  period");
      Put_Line ("   switch can be :");
      Put_Line ("            -u get this help");
      Put_Line ("            -v verbose mode ");
      New_Line;
   end Usage;

begin

   Copyright ("schedule_hardcoded_model");

   -- Get arguments
   --
   loop
      case GNAT.Command_Line.Getopt ("u v ") is
         when ASCII.NUL =>
            exit;
         when 'v' =>
            Verbose := True;
         when 'u' =>
            Usage;
            OS_Exit (0);
         when others =>
            Usage;
            OS_Exit (0);
      end case;
   end loop;

   loop
      declare
         S : constant String :=
           GNAT.Command_Line.Get_Argument (Do_Expansion => True);

      begin
         exit when S'Length = 0;
         Period_String := Period_String & S;
      end;
   end loop;

   -- Check the period on which we will compute the scheduling
   --
   to_natural (Period_String, Period, Ok);
   if not Ok then
      Raise_Exception
        (Constraint_Error'Identity,
         "The period to compute the scheduling must be a numeric value");
   end if;

   -- Initialize the Cheddar framework
   --
   Call_Framework.initialize (False);

   -- Build architecture model
   --
   test1_xml (Sys);

   -- Compute the scheduling on the period given by the argument
   --
   Initialize (Response_List);
   Initialize (Request_List);
   Initialize (A_Request);
   A_Request.statement    := Scheduling_Simulation_Time_Line;
   A_Param                := new Parameter (Integer_Parameter);
   A_Param.parameter_name := To_Unbounded_String ("period");
   A_Param.integer_value  := Period;
   Add (A_Request.param, A_Param);
   A_Param                := new Parameter (Boolean_Parameter);
   A_Param.parameter_name := To_Unbounded_String ("schedule_with_offsets");
   A_Param.boolean_value  := True;
   Add (A_Request.param, A_Param);
   A_Param                := new Parameter (Boolean_Parameter);
   A_Param.parameter_name :=
      To_Unbounded_String ("schedule_with_precedencies");
   A_Param.boolean_value  := True;
   Add (A_Request.param, A_Param);
   A_Param                := new Parameter (Boolean_Parameter);
   A_Param.parameter_name := To_Unbounded_String ("schedule_with_resources");
   A_Param.boolean_value  := True;
   Add (A_Request.param, A_Param);
   A_Param                := new Parameter (Integer_Parameter);
   A_Param.parameter_name := To_Unbounded_String ("seed_value");
   A_Param.integer_value  := 0;
   Add (A_Request.param, A_Param);

   for I in Time_Unit_Event_Type'Range loop
      A_Param                := new Parameter (Boolean_Parameter);
      A_Param.parameter_name := to_lower (I'Img);
      A_Param.boolean_value  := True;
      Add (A_Request.param, A_Param);
   end loop;

   Add (Request_List, A_Request);
   Sequential_Framework_Request (Sys, Request_List, Response_List);

   for J in 0 .. Response_List.Nb_Entries - 1 loop
      Put_Line (To_String (Response_List.Entries (J).title));
      Put_Line (To_String (Response_List.Entries (J).text));
   end loop;

   for i in 0 .. Sched.nb_entries - 1 loop
      if Sched.entries (i).data.error_msg /= empty_string then
         Put_Line
           ("Error_Msg for processor " &
            To_String (Sched.entries (i).item.name) &
            " : " &
            To_String (Sched.entries (i).data.error_msg));
      end if;
   end loop;

   Write_To_Xml_File
     (framework.Sched,
      Sys,
      To_Unbounded_String ("schedule_hardcoded_result.xml"));

exception

   when others =>
      Put_Line ("Exception name    : " & Exception_Name);
      Put_Line ("Exception message : " & Exception_Message);

end schedule_hardcoded_model;
